--[[
**********************************************************************
MagicDKP - a raid loot and DKP tracker.
**********************************************************************
This file is part of MagicDKP, a World of Warcraft Addon

MagicDKP is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MagicDKP is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MagicDKP.  If not, see <http://www.gnu.org/licenses/>.
**********************************************************************
]]

--[[ Config dummy

_[Settings]_[Items]_[Players]_[Raids]_[Events]_________


Items =>
   Item Name
     Cost => Item Cost
     Type => Ignore / Bank / Fixed Cost / Variable Cost
     Item Link => Link

Players =>
  Name =>
    Level
    Class
    Guid
    Race
    Alts => [a, b, c]

Raids =>
  RaidDate/Name =>
    Start, End => Start/end of raid
    OnTime =>
       attendees = { PlayerName },
       standby = {}
    BossAttempt =>

 ]]

local mod = MagicDKP
local L = LibStub("AceLocale-3.0"):GetLocale("MagicDKP", false)
local R = LibStub("AceConfigRegistry-3.0")
local C = LibStub("AceConfigDialog-3.0")
local DBOpt = LibStub("AceDBOptions-3.0")

local BabbleZone = LibStub("LibBabble-Zone-3.0") 
local ZR = BabbleZone:GetReverseLookupTable()
local ZL  = BabbleZone:GetLookupTable()

local BabbleBoss = LibStub("LibBabble-Boss-3.0") 
local BR = BabbleBoss:GetReverseLookupTable()
local BL = BabbleBoss:GetUnstrictLookupTable()

local LDBIcon = LibStub("LibDBIcon-1.0", true)
local LDB = LibStub("LibDataBroker-1.1", true)


local fmt = string.format

local options, cmdoptions, standardBossOptions, standardEventOptions, standardRaidOptions, stdEventOptions, playerOptions
local logLevelsDropdown = {
   [0] = L["Disabled"], 
   [1] = L["Errors only"], 
   [2] = L["Errors and warnings"], 
   [3] = L["Informational messages"], 
   [4] = L["Debug messaging"], 
   [5] = L["Debug trace messages"], 
   [6] = L["Highest level of debug log spam"], 
}
local zones, db, items, events, raids, players 

local idmatch = "(%a+):(%d+)"


BabbleZone = nil

local function UpdateUpvalues()
   db = mod.db.profile         -- profile data
   zones = MagicDKPDB.zones     -- raid zones / boss mobs
   items = MagicDKPDB.items     -- item database
   events = MagicDKPDB.events   -- custom events
   raids = MagicDKPDB.raids     -- raids
   players = MagicDKPDB.players -- player db
end

local playerInfoString = L["Player: |cff33ff99%s|r"].."\n"..L["Class: |cff33ff99%s|r"].."\n"..L["Level: |cff33ff99%d|r"].."\n"..L["Rank: |cff33ff99%s|r"].."\n"..L["Main Character: |cff33ff99%s|r"]
local function GetPlayerInfo(var)
   local player = var[#var-1]
   local db = players[player]
   
   return playerInfoString:format(player, db.class, db.level, db.rank, db.main and ("No (".. db.main..")") or "Yes")
end

function mod:MultiDKPDisabled()
   return not db.enableMultiDKP
end

do
   local dkppools = {}
   function mod:GetMultiDKPPoolList()
      for id in pairs(dkppools) do dkppools[id] = nil end
      local id = 1
      dkppools["D"] = L["Default"]
      local DB = mod:Database()
      if DB then
	 local pools = DB:GetMultiDKPPools()
	 if pools then
	    for name, val in pairs(pools) do
	       dkppools[name] = val
	    end
	 end
      end
      return dkppools
   end
end

local function GetPlayerItems(var)
   local player = var[#var-1]
   local DB = mod:Database()
   if DB then 
      local data = DB:GetPlayerItems(player)
      local result = {}
      if not data then
	 local main = players[player].main
	 if main then
	    data = DB:GetPlayerItems(main)
	 end
      end
      if data then
	 local result = {}
	 for _, item in ipairs(data) do
	    local name = item.name
	    if item.itemid then
	       name = (select(2, GetItemInfo(item.itemid))) or item.name or ("[item:%d]"):format(item.itemid)
	    end
	    result[#result+1] = name..":  |cff33ff99"..item.dkp.."|r"
	 end
	 return table.concat(result, "\n")
      end
   end
   return L["No item history found."]
end

local function GetPlayerDKP(var)
   local player = var[#var-1]
   local DB = mod:Database()
   if DB then
      local data =  DB:GetPlayerData(player)
      local result = {}
      if not data then
	 local main = players[player].main
	 if main then
	    data = DB:GetPlayerData(main)
	    if data then
	       result[1] = fmt(L["No data found for %s. Showing data for main character %s.\n"], player, main)
	    end
	 end
      end
      if data then 
	 for key,val in pairs(DB:GetMultiDKP()) do
	    result[#result+1] = key.." "..L["Available"]..":  |cff33ff99"..data[val.current].."|r"
	    result[#result+1] = key.." "..L["Earned"]..":  |cff33ff99"..data[val.earned].."|r"
	    result[#result+1] = key.." "..L["Spent"]..":  |cff33ff99"..data[val.spent].."|r"
	    result[#result+1] = key.." "..L["Adjustments"]..":  |cff33ff99"..data[val.adjust].."|r"
	    result[#result+1] = ""
	 end
	 return table.concat(result, "\n")
      end
   end
   return L["No DKP history found."]
end


local function GetBossName(arg)
   local region = arg[#arg-2]
   local name =  zones[ region ] and zones[ region ].mobs[ tonumber(arg[#arg-1]) ].name
   return name and  BL[name] or name   
end

local function highlite(str)
   return fmt("|cff33ff99%s|r", tostring(str))
end

local function GetItemDesc(arg)
   local itemid
   local var,num
   if type(arg) == "table" then
      var, num = string.match(arg[#arg], idmatch)
   end
   itemid = (num and tonumber(num)) or (type(arg) == "table" and tonumber(arg[#arg-1])) or tonumber(arg)
   return itemid..":"..items[itemid].itemInfo[2]
end

local function GetEventName(info)
   return info[#info-1]
end
   
local function GetEventDesc(info)
   local event = info[#info-1]
   local raid = tonumber(info[#info-2])

   return mod:FromUnixTime(raids[raid].events[event].eventTime, L["Event Time"]..": |cffffffff", "|r")
end

local function GetRaidName(raidid)
   local name = raids[raidid].zone
   return mod:FromUnixTime(endTime and raids[raidid].endTime or raids[raidid].startTime,
			   nil, name and ZL[name], name)
end


local function GetRaidDesc(info)
   local raid = tonumber(info[#info-1])
   local desc
   desc = mod:FromUnixTime(raids[raid].startTime, L["Raid Start Time"]..": |cffffff50", "|r")
   if raids[raid].endTime then
      desc = desc .. "\n"..mod:FromUnixTime(raids[raid].endTime, L["Raid End Time"]..": |cffffff50", "|r")
   else
      desc = desc .."\n".. L["Raid End Time"]..": |cffffff50".. L["Still in progress"].. "|r"
   end
   return desc
end

local function GetItemTexture(arg)
   local var,num = string.match(arg[#arg], idmatch)
   local itemid = tonumber(arg[#arg-1])
   if var == "name" and num then
      itemid = tonumber(num)
   end
   return items[itemid].texture
end


local function GetItemBidHistory(arg)
   local itemid = tonumber(arg[#arg-1])
   local DB = mod:Database()
   if DB then
      local history = DB:GetItemBidHistory(itemid)
      if #history then
	 local fmt = "%s: %d\n"
	 local desc = ""
	 for _,data in ipairs(history) do
	    desc = desc..fmt:format(data.name, data.dkp)
	 end
	 return desc
      end
   end
   return L["No bids found."]
end



function mod:ListAlts(_,alt)
   local alts = {}
   if alt then
      alt = alt:lower():trim():gsub("^.", string.upper)
   end
   if not alt or not players[alt] then 
      if self.error then self:error("Invalid parameters: Expected a valid player name.") end
      return
   end
   for name,data in pairs(players) do
      if data.main == alt then
	 alts[#alts+1] = name
      end
   end
   if #alts > 0 then
      table.sort(alts)
      self:Print(fmt("Alts registered for %s: %s", alt, table.concat(alts, ", ")))
   else
      self:Print(fmt("There are no alts registered for %s", alt))
   end
end

function mod:AssociateAlt(alt, player)
   if not player or not alt then
      if self.error then self:error("Invalid parameters: Expected [alt player].") end
      return
   end
   alt = alt:lower():trim():gsub("^.", string.upper)
   player = player:lower():trim():gsub("^.", string.upper)
   if not players[player] then
      if self.error then self:error("Unknown player: %s", player) end
      return
   end
   if not players[alt] then
      self:UpdatePlayerInfo(alt, "Unknown", 0, "Unknown")
   end
   if players[player].main then
      if self.error then self:error("Player %s is already registered as an alt for %s.", player, players[player].main) end
      return
   end
      
   players[alt].main = player
   if self.hasInfo then self:info("Registered <%s> as an alt for <%s>.",  alt, player) end
end

local cmdoptions = {
   type = "group",
   name = L["Magic DKP"],
   handler = MagicDKP,
   args = {
      additem = {
	 type = "input",
	 name = L["args: <looter> <itemlink> - add an item to the current raid/event"],
	 set = "AddItemManually",
	 get = function()end,
      },
      bid = {
	 type = "input",
	 name = L["args: <itemlink> - Start a new DKP bid"],
	 set = "StartDKPBid",
	 get = function()end,
      },
      alt = {
	 type = "group",
	 name = L["Alternate character management."],
	 args = {
	    add = {
	       type = "input",
	       name = L["alt add <alt> <player> - associate <alt> to <player>"],
	       set = function(info, val) mod:AssociateAlt(strsplit(" ", val)) end,
	       get = function() end,
	    },
	    list = {
	       type = "input",
	       name =  L["alt list <player> - list all alts of <player>"],
	       set = "ListAlts",
	       get = function() end,
	    },
	    del = {
	       type = "input",
	       name = L["alt clear <alt> - remove the alt association for <alt>"],
	       set = function(info, val)
			if val then
			   val = val:lower():trim():gsub("^.", string.upper)
			end			
			if players[val] then
			   players[val].main = nil
			   if mod.info then mod:info("%s is no longer registered as an alt.", val) end
			else
			   if mod.error then mod:error("Unknown player: %s", val) end
			end
		     end,
	       get = function() end,
	    },
	 }
      },
      ["end"] = {
	 type = "execute",
	 name = L["Force the current active raid, if any, to end."],
	 func = function()
		   if not db.enabled then mod:Print("Addon is not enabled. Use /magicdkp toggle to enable it.") return end
		   if not mod:IsLootMaster() then mod:Print("You are not the master looter / raid leader, not ending raid.") return end
		   mod:EndRaid()
		   mod:UnregisterEvent("ZONE_CHANGED_NEW_AREA")
		   mod:UnregisterEvent("PLAYER_REGEN_DISABLED")
		   mod:UnregisterEvent("UPDATE_MOUSEOVER_UNIT")   
		   mod:UnregisterEvent("PLAYER_TARGET_CHANGED")
		   mod:UnregisterEvent("CHAT_MSG_LOOT")
		   mod:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		end
      },
      start = {
	 type = "execute",
	 name = L["Force the current raid to end and a new one to be created."],
	 func = function()
		   if not db.enabled then mod:Print("Addon is not enabled. Use /magicdkp toggle to enable it.") return end
		   if not mod:IsLootMaster() then mod:Print("You are not the master looter / raid leader, not starting event.") return end
		   mod:EndRaid() -- end old raid if present
		   mod:FindOrCreateRaid() -- create a new one
		   mod:RegisterEvent("UPDATE_MOUSEOVER_UNIT", "ScanTarget", "mouseover")   
		   mod:RegisterEvent("PLAYER_TARGET_CHANGED", "ScanTarget", "target")   
		   mod:RegisterEvent("ZONE_CHANGED_NEW_AREA")
		   mod:RegisterEvent("CHAT_MSG_LOOT")	    
		   mod:RegisterEvent("PLAYER_REGEN_DISABLED")
		   mod:ZONE_CHANGED_NEW_AREA() -- trigger zone change logic to run
		end,
      },
      toggle = {
	 type = "toggle",
	 name = L["Enable or disable the addon."],
	 set = function()
		  db.enabled = not db.enabled
		  if db.enabled then
		     mod:OnEnable()
		     mod:Print("Addon is now enabled.")
		  else
		     mod:OnDisable()
		     mod:Print("Addon is now disabled.")
		  end
	       end,
	 get = function() return db.enabled end,
      },
      versions = {
	 type = "execute",
	 name = L["Query raid for their MagicDKP versions."],
	 func = "QueryAddonVersions",
      },
      config = {
	 type = "execute",
	 name = L["Toggle configuration dialog"],
	 func = "ToggleConfigDialog",
      },
      wipe = {
	 type = "execute",
	 name = L["Log a wipe for the current boss"],
	 func = "AddWipe",
      },
      event = {
	 type="group",
	 name = L["Take a custom event snapshot."],
	 args = {
	 }
      }
      
   }
}
	 
local options = {
   type = "group", 
   name = L["Magic DKP"],
   childGroups = "tab",
   handler = MagicDKP,
   args = {
      settings = {
	 type = "group",
	 name = L["Settings"],
	 set = "SetProfileParam",
	 get = "GetProfileParam",
	 args = {
	    clneanup = {
	       type = "group",
	       name = L["Data Housekeeping"],
	       order = 100,
	       args = {
		  sep0 = {
		     type = "header",
		     name = L["Raid Expiration"],
		     order = 5,
		     width = "full",
		  },
		  maxraidage = {
		     type = "select",
		     name = L["Max Raid Age"],
		     values = {
			L["Never Delete"],
			L["6 Months"],
			L["3 Months"],
			L["1 Month"]
		     },
		     order = 10,
		  },
		  expireraids = {
		     type = "execute",
		     name = L["Delete old raids"],
		     func = "ExpireOldRaids",
		     order = 20,
		  },
		  sep1 = {
		     type = "header",
		     name = L["Item Cleanup"],
		     order = 25,
		     width = "full",
		  },
		  deleteIgnored = {
		     type = "toggle",
		     name = L["Delete Ignored Items"],
		     desc = L["When deleting unused items, also delete ignored items."], 
		     order = 30,
		  },
		  deleteBank = {
		     type = "toggle",
		     name = L["Delete Banked Items"],
		     desc = L["When deleting unused items, also delete items that are sent to the bank by default."], 
		     order = 30,
		  },
		  expireitems = {
		     type = "execute",
		     name = L["Delete Unused Items"],
		     desc = L["Delete all items not currently used in any raids. Note that predefined items such as badges will be kept. Unless specifically selected, ignored and banked items will be not be removed."],
		     func = "ExpireOldItems",
		     order = 40
		  },
		  sep2 = {
		     type = "header",
		     name = L["Player Cleanup"],
		     order = 45,
		     width = "full",
		  },
		  expireOOGPlayer = {
		     type = "execute",
		     name = L["Expire out-of-guild players."],
		     desc = L["Delete all players not in your current guild. Players in existing raids will also be kept however."],
		     order = 50,
		     func = "ExpireOOGPlayers", 
		  }
	       },
	    },
	    multidkp = {
	       type = "group",
	       name = L["Multi DKP"],
	       order = 1,
	       args = {
		  help = {
		     type = "description",
		     name =
			L["MultiDKP is a feature in EQDKPPlus that allows tracking of multiple separate DKP pools."].." "..
			L["MagicDKP supports zone and item overrides for MultiDKP."].." "..
			L["Note that MultiDKP support requires a valid GetDKP import in the DKP Database."].." "..
			L["The default DKP Pool to use when there is no zone or item override is configured under the DKP Database."],
		     order = 0
		  },
		  enableMultiDKP = {
		     type = "toggle",
		     name = L["Enable MultiDKP Support"],
		     desc = L["Enable zone and item level overrides for MultiDKP tracking. If disabled, the default dkp pool as configured in the DKP Database will be used."],
		     width = "full", 
		  },
		  
	       },
	       
	    },
	    messaging = {
	       type = "group",
	       name = L["Notifications"],
	       order = 1,
	       args = {
		  lootAwarded = {
		     type = "input", 
		     name = L["Award Message"],
		     desc = L["The message to send to the raid when awarding an item."],
		     width = "full",
		     hidden = function() return not db.lootNotification end,
		     order = 21
		  },
		  lootNotification = {
		     type = "toggle", 
		     name = L["Enable loot notification messages"],
		     desc = L["When enabled, a message is sent to guild chat and/or raid chat when an item is awarded"],
		     width = "full",
		     order = 22
		  },
		  notifyGuild = {
		     type = "toggle", 
		     name = L["Send loot notification to guild chat"],
		     desc = L["Send the loot notification message to guild chat instead of raid chat."],
		     width = "full",
		     order = 22
		  },
		  reportraid = {
		     type = "toggle", 
		     name = L["Report events taken to raid chat"],
		     desc = L["Send a message to raid chat when an event is recorded."],
		     width = "full",
		     order = 26
		  },
		  reportstandby = {
		     type = "toggle", 
		     width = "full",
		     name = L["Report standby events to guild chat"],
		     desc = L["Send a message to guild chat when recording a standby event."],
		     order = 26
		  },		  
	       }
	    },
	    standby = {
	       type = "group",
	       name = L["Alts and Standby"],
	       order = 1, 
	       args = {
		  detectAlts = {
		     type = "toggle",
		     width="full",
		     name = L["Attempt to detect alts in the guild notes"],
		     desc = L["If enabled, MagicDKP will attempt to automatically discover the main character of alternate characters by looking looking in first the guild note, then the officer note, for text matching the following pattern: |cffc9ffa9Alt: /main char name/|r"],
		     order = 25
		  },
		  clearAlt = {
		     type = "toggle",
		     width="full",
		     name = L["Clear alt when guild note is missing"],
		     desc = L["If enabled, remove alts from any character where the alt guild note is not discovered. Note that this might not work correctly for out of guild alts."],
		     order = 26
		  },
		  standbyTracking = {
		     type = "toggle",
		     width="full",
		     name = L["Enable standby tracking"],
		     desc = L["When enabled MagicDKP sends a standby check to all guild members of the ranks listed below that are not in the raid."],
		     order = 25
		  },
		  header = {
		     type = "header",
		     name = L["Standby Guild Ranks"],
		     order = 100
		  },
		  help = {
		     type = "description",
		     name = L["Select the raider ranks of your guild. Only the selected ranks will be included in the raid standby calls."],
		     order = 101,
		  },
		  ranks = {
		     type = "multiselect",
		     name = "",
		     values = "GetGuildRanks", 
		     get = "GetGuildRank",
		     set = "SetGuildRank",
		     order = 102
		  }		  		  
	       }
	    }, 
	    exporting = {
	       type = "group",
	       name = L["XML Exports"],
	       order = 1, 
	       args = {
		  standbyattendee = {
		     type = "toggle",
		     name = L["Add Standby to Raid"],
		     desc = L["If checked, standby attendees will be added to the regular list of attendees for an event when doing a CTRT export."],
		     order = 25
		  },
		  eventnameaszone = {
		     type = "toggle",
		     name = L["Use custom event name as zone name"],
		     desc = L["If checked, the name of the custom event will be used for both the zone name and boss kill name when exporting to CTRT. If unchecked, the zone name will be the name of the raid level zone."],
		     width = "full",
		     order = 27
		  },
	       }	       
	    },
	    options = {
	       type = "group",
	       name = L["General Options"],
	       order = 1,
	       args = {
		  enabled = {
		     type = "toggle",
		     name = L["Enable MagicDKP event tracking."],
		     desc = L["When the addon is disabled, it will not allow raids to be created or enable by itself. It will also not listen to loot events and such."],
		     width = "full",
		     set = function()
			      db.enabled = not db.enabled
			      if db.enabled then
				 mod:OnEnable()
				 mod:Print("Addon is now enabled.")
			      else
				 mod:OnDisable()
				 mod:Print("Addon is now disabled.")
			      end
			   end,
		  },
		  trackonly = {
		     type = "toggle", 
		     width = "full",
		     name = L["Only track loot, don't ask for DKP values"],
		     desc = L["When recording loot, never ask for a bid value but add an entry. Useful if you done use DKP and just want to keep track of who gets what items."],
		     order = 29
		  },
		  logLevel = {
		     type = "select",
		     name = L["Log level"],
		     desc = L["The logging level determines the amount of output printed by the addon. Debug can be useful as you're getting use to the addon or want to figure out why it marked in a specific way. Trace is only useful for debugging and for development purposes."],
		     values = logLevelsDropdown,
		     order = 10,
		  },
		  raidsize  = {
		     type = "select",
		     name = L["Minimum Raid Size"],
		     desc = L["Only enable tracking when you're part of a raid larger than this. Note that tracking only will be enabled if you're actually in a raid, regardless of the value of this option."], 
		     values = {
			["25"] = L["25-Man"],
			["10"] = L["10-Man"],
			["2"]  = L["Any size raid"],
		     },
		     order = 20,
		     
		  },
		  trackwipe = {
		     type = "toggle",
		     name = L["Track wipes as events"],
		     desc = L["Whether or not MagicDKP should track boss attempt wipes. If a boss is marked as being on farm status, no wipe tracking will be done."],
		     order = 30,
		     hidden = true,
		  },
		  xloothook = {
		     width="full",
		     type = "toggle",
		     name = L["Enable XLootMaster hook"],
		     desc = L["When enabled, MagicDKP will hook into the XLootMaster addon to report loot events. This allows MagicDKP to register looting even when a player is out of range. Note that this might mean that an item is added without the player receiving it (for example due to a full inventory). In those instances MagicDKP will try to do duplicate detection and you'll be asked if you want to add another item for the event."],
		     hidden = function() return IsAddOnLoaded("XLootMaster") ~= 1 end
		  },
		  external = {
		     width="full",
		     type = "toggle",
		     name = L["Enable MagicLooter support"],
		     desc = L["When enabled, MagicDKP will listen to loot distribution events from MagicLooter Loot Menu."],
		  },
		  lootEventListen = {
		     width="full",
		     type = "toggle",
		     name = L["Listen to Loot Events"],
		     desc = L["When enabled, MagicDKP will listen to loot events. If you're using an external loot notification (i.e the XLootMaster hook, MagicDKP Bidder or MagicLooter LootMenu), this is not needed nor recommended."],
		  },
		  logPlayerEventData = {
		     width="full",
		     type = "toggle",
		     name = L["Log Joins and Leaves"],
		     desc = L["Enabled logging of players joining or leaving the raid. The resulting data can be used to enable time based DKP tracking. Note that this data is currently only surfaced in the MagicDKP export format."],
		  },
		  wipepercentage = {
		     type = "range",
		     name = L["Wipe Percentage"],
		     desc = L["The percentage of the raid that has to be dead before MagicDKP asks whether to record a wipe event."],
		     min = 1, max = 100, step = 1,
		     order = 40,
		     hidden = true,
		     disabled = function() return not db.trackwipe end,
		  },
		  minimapIcon = {
		     type = "toggle",
		     name = L["Enable minimap icon"],
		     desc = L["Show an icon to open the config at the Minimap"],
		     width="full",
		     get = function() return not db.minimapIcon.hide end,
		     set = function(info, value) db.minimapIcon.hide = not value; LDBIcon[value and L["Show"] or L["Hide"]](LDBIcon, "MagicDKP") end,
		     disabled = function() return not (LDBIcon and LDB)  end,
		  },
		  
	       }
	    },
	 }, 
	 order = 1,
      }, 
      zones = {
	 type = "group",
	 childGroups = "select",
	 name = L["Raid Zones"],
	 handler = mod, 
	 args = {}, 
	 order = 2,
      }, 
      events = {
	 type = "group",
	 name = L["Events"],
	 set = "SetEventName",
	 get = "GetEventName",
	 args = {
	    addcustom = {
	       type = "group",
	       name = L["Add custom event"],
	       args = {
		  eventname = {
		     type = "input",
		     name = L["Event name"],
		     order = 1,
		     set = "SetAddCustomEvent",
		     get = "GetAddCustomEvent",
		  },
		  addevent = {
		     type = "execute",
		     name = L["Add event"],
		     order = 1000,
		     func = "AddCustomEvent"
		  },
	       },
	       order = 100000,
	    },
	 }, 
	 order = 3,
      }, 
      itemdb = {
	 type = "group",
	 name = L["Item database"],
	 args = {
--	    add = {
--	       type = "group",
--	       name = L["Add Item..."],
--n	       order = 0,
--	       args = {
--	       }
--	    }
	 }, 
	 order = 4,
      }, 
      playerdb = {
	 type = "group",
	 name = L["Player database"],
	 args = {}, 
	 order = 5,
      }, 
      raids = {
	 type = "group",
	 name = L["Raids"],
	 args = {}, 
	 order = 6,
      }, 
   }
}

playerOptions = {
   description = {
      name = GetPlayerInfo,
      type = "description",
      order = 0,
   },
   footnote = {
      name = "\n\n"..L["Note: Use the  |cff33ff99/md alt|r commands to edit alts. UI editing TBD."],
      type = "description",
      order = 1,
   },
   dkpHeader = {
      type = "header",
      name = L["Player DKP"],
      hidden = "NoDKPDatabase",
      order = 10,
   }, 
   dkp = {
      name = GetPlayerDKP,
      type = "description",
      hidden = "NoDKPDatabase",
      order = 20,
   },
   itemHeader = {
      type = "header",
      name = L["Item History"],
      hidden = "NoDKPDatabase",
      order = 30,
   }, 
   items = {
      name = GetPlayerItems,
      type = "description",
      hidden = "NoDKPDatabase",
      order = 40,
   }
   
}

standardBossOptions = {
   header = {
      name = GetBossName,
      type = "header",
      order = 0
   },
   isfarm = {
      name = L["Is farm"],
      type = "toggle",
      desc = L["If a mob is on farm mode, MagicDKP will not ask about wipe DKP events when wiping on the boss."],
      order = 2,
      hidden = true, 
   },
   ignore = {
      name = L["Ignore boss"],	
      type = "toggle",
      desc = L["Don't ask to take attendance when this boss dies."],
      order = 1, 
   },
   event = {
      name = L["Event name"],
      type = "input",
      desc = L["Name of the event to use when adding the even to the current raid, i.e boss name or encounter name such as Illidari Council or Reliquary of Souls. If multiple non-ignored mobs share the same event, MagicDKP will not ask to take attendance until all the mobs are dead. When you modify the event name, all boss mobs with the same event will be modified."],
      order = 10,
   }, 
   resetevent = {
      name = L["Reset event name"],
      type = "execute",
      desc = L["Set the event name to the default value. This will NOT affect other boss mobs with the sane event name."],
      order = 12,
      func = "ResetBossEvent",
      hidden = "IsDefaultBossEvent",
   }, 
   notes = {
      name = L["Notes"],
      type = "input",
      multiline = true,
      order = 100,
      width = "full",
   }
}

standardEventOptions = {
   header = {
      name = GetEventName,
      type = "header",
      order = 0
   },
--   value = {
--      name = L["Event Value"],
--      type = "input",
--      desc = ["EVENT VALUE HELP"],
--      order = 10,
--   }, 
   eventInfo = {
      type = "description",
      name = GetEventDesc,
      order = 2
   },
   lootheader  = {
      type = "header",
      name = L["Loot"],
      order = 10000
   }
   
}

standardRaidOptions = {
   header = {
      name = function(info) return GetRaidName(tonumber(info[#info-1]))end,
      type = "header",
      order = 0
   },
   raidInfo = {
      type = "description",
      name = GetRaidDesc,
      order = 2
   },
   export = {
      type = "execute",
      name = L["CTRT Export"],
      desc = L["Export raid information in a format understood by CT_RaidTracker compatible importers."],
      func = "ExportCTRT",
   },
   exportmdkp = {
      type = "execute",
      name = L["MagicDKP Export"],
      desc = L["Export raid information in MagicDKP's own XML format."],
      func = "ExportMagicDKP",
   }
}


stdEventOptions = {
   type = "group",
   name = function(info) return info[#info] end,
   args = {
      eventname = {
	 type = "input",
	 name = L["Event name"],
	 order = 1
      },
      delevent = {
	 type = "execute",
	 name = L["Delete"],
	 order = 2,
	 func = "DeleteEvent",
      },
   }
}

local cmdEventOption = {
   type = "execute",
   name = L["Take attendance for this event"],
   func = function(info) mod:LogCustomEvent(info[#info]) end
}

standardItemOptions = {
   iteminfo = {
      type = "input",
      name = GetItemTexture,
      dialogControl = "MDKPItemWidget",
      width = "full",
      order = 0,
      get = GetItemDesc,
   },
   ignore = {
      name = L["Ignore"],	
      type = "toggle",
      desc = L["When an item is ignored, MagicDKP will not log any information about it when it's looted."],
      order = 1, 
   },
   bank = {
      name = L["Bank Item"],
      type = "toggle",
      desc = L["Bank items will be assigned to a special bank character with a zero value. Use this for items you want to track but which should have no recipients, for example epic gems."],
      disabled = "IsItemIgnored",
      order = 2, 
   }, 
   type = {
      name = L["Fixed Value"],
      type = "toggle",
      desc = L["If an item has a fixed value MagicDKP will not ask for a value when the item is looted. It will use the value configured on this screen."],
      disabled = "IsItemUnvalued",
      order  = 4,
   }, 
   value = {
      name = L["Value"],
      type = "input",
      desc = L["The default or fixed value of the item. Bank items will always be logged with a value of zero. For bidding DKP this value is just used as the default."],
      disabled = "IsItemUnvalued",
      order = 3,
   },
   multiDKP = {
      name = L["MultiDKP Pool"],
      type = "select",
      desc = L["The DKP Pool to use when taking bids and charging for this item. This overrides both the default and zone-wide setting. If left at Default, it will use whatever the tier defined for the players zone or the overall data."],
      hidden = "MultiDKPDisabled",
      values = "GetMultiDKPPoolList",
      order = 4
   },
   bidhistoryheader = {
      name  = "Bid History",
      type = "header",
      order = 10000,
      hidden = "NoDKPDatabase", 
   },
   bidhistory = {
      name  = GetItemBidHistory,
      type = "description",
      hidden = "NoDKPDatabase",
      order = 10001
   },
   
}

do
   local tmpCustomEvent
   function mod:SetAddCustomEvent(info,val) tmpCustomEvent = val end
   function mod:GetAddCustomEvent(info) return tmpCustomEvent end
   function mod:AddCustomEvent(info)
      if not tmpCustomEvent then return end
      local key = self:MakeKey(tmpCustomEvent)
      if strlen(key) > 0 then
	 key = strlower(key)
	 events[key] = tmpCustomEvent
	 options.args.events.args[key] = stdEventOptions
	 cmdoptions.args.event.args[key] = cmdEventOption
	 self:NotifyChange()
      end
      tmpCustomEvent = nil
   end

   function mod:DeleteEvent(info)
      local key = info[#info-1]
      events[key] = nil
      options.args.events.args[key] = nil
      cmdoptions.args.event.args[key] = nil
   end
end


do 
   local simpleNameCache = {}
   function mod:MakeKey(name)
      if not name then return "" end
      if not simpleNameCache[name] then
	 simpleNameCache[name] = gsub(name, " ", "")
      end
      return simpleNameCache[name]
   end
end


function mod:GetGuildRanks()
   local ranks = {}
   for id = 1, GuildControlGetNumRanks() do
      ranks[fmt("%02d", id)] = fmt("%d: %s", id, GuildControlGetRankName(id))
   end
   return ranks
end

function mod:GetGuildRank(info, key)
   key = tonumber(key)
   if db.ranks then
      return db.ranks[key]
   end
end

function mod:SetGuildRank(info, key, state)
   key = tonumber(key)
   if not db.ranks then db.ranks = {} end
   if state then
      db.ranks[key] = state
   else
      db.ranks[key] = nil
   end
end

function mod:AddPlayerConfig(player)
   local hash = options.args.playerdb.args[player] or {}
   hash.name = player
   hash.type = "group"
--   itemcfg.get = "GetOption"
--   itemcfg.set = "SetOption"
   hash.args = playerOptions

   options.args.playerdb.args[player] = hash
end

function mod:BuildPlayerConfig()
   local hash = options.args.playerdb.args
   for player, data in pairs(players) do
      mod:AddPlayerConfig(player)
   end
   self:NotifyChange()
end

function mod:BuildZoneConfig(var)
   local zone = (type(var) == "table" and var[#var-1]) or var
   local zoneHash = options.args.zones.args[zone]

   if self.hasTrace then self:trace("Generating configuration for mobs in zone %s", zoneHash.name) end

   zoneHash.args.loader.hidden = true
   
   for mob,data in pairs(zones[zone].mobs) do
      if self.hasSpam then self:spam("  Added %s...", data.name) end
      mob = tostring(mob)
      if not zoneHash.args[mob] then 
	 zoneHash.args[mob] = {
	    type = "group",
	    name = BL[data.name] or data.name,
	    args =  standardBossOptions,
	    get = "GetBossOption",
	    set = "SetBossOption",
	    --	 plugins =  {
	    --	    extras = {
	    --	       type = 'execute',
	    --	       name = L['Delete mob from database (not recoverable)'],
	    --	       order = 10001,
	    --	       width = "full",
	    --	       func = "RemoveMob",
	    --	       confirm = true,
	    --	       confirmText = fmt(L["Are you sure you want to delete |cffd9d919%s|r from the database?"], BL[data.name])
	    --	    }
	    --      }
	 }
      end
   end
   self:NotifyChange()
end
 
local tmpsorttable = {}
local function GetSortedList(attend)
   for id in pairs(tmpsorttable) do
      tmpsorttable[id] = nil
   end
   if attend then
      for id in pairs(attend) do
	 local pinfo = players[id] 
	 if pinfo and pinfo.main then
	    id = fmt("%s (%s)", pinfo.main, id)
	 end
	 tmpsorttable[#tmpsorttable+1] = id
      end
      table.sort(tmpsorttable)
      return table.concat(tmpsorttable, ", ")
   end
   return "None"
end
local function GetRaidAttendees(info)
   local event = info[3]
   local raidid = tonumber(info[2])
   if mod.spam then mod:spam("GetRaidAttendees(%s, %s)", tostring(raidid), tostring(event)) end
   return GetSortedList(raids[raidid].events[event].attendees)
end

local function GetRaidStandby(info)
   local event = info[3]
   local raidid = tonumber(info[2])
   if mod.spam then mod:spam("GetRaidStandby(%s, %s)", tostring(raidid), tostring(event)) end
   return GetSortedList(raids[raidid].events[event].standby)
end

local attendeeConfig = {
   attendHeader = {
      type = "header",
      name = L["Attendees"],
      order = 50,
   },
   attendList = {
      type = "description",
      name = GetRaidAttendees,
      order = 51,
   },
   addAttendeeInput = {
      type = "input",
      name = L["Add extra attendee"],
      order = 52,
      set = "AddExtraAttendee",
   },
   delAttendeeInput = {
      type = "input",
      name = L["Delete attendee"],
      order = 52,
      set = "DelExtraAttendee",
   },
   
   standbyHeader = {
      type = "header",
      name = L["Standby Attendees"],
      order = 100,
   },
   standbyList = {
      type = "description",
      name = GetRaidStandby,
      order = 101,
   },
   addStandbyInput = {
      type = "input",
      name = L["Add extra standby attendee"],
      order = 102,
      set = "AddExtraAttendee",
   },
   delStandbyInput = {
      type = "input",
      name = L["Delete standby attendee"],
      order = 102,
      set = "DelExtraAttendee",
   },
}

local function cap(str)
   return str:lower():trim():gsub("^.", string.upper)
end


function mod:AddExtraAttendee(var,val)
   local type = var[#var]
   local raidid = tonumber(var[#var-2])
   local event = var[#var-1]
   local hash
   val = cap(val)
   if type == "addStandbyInput" then
      type = "standby"
      hash = raids[raidid].events[event].standby or {}
      raids[raidid].events[event].standby = hash
   else
      type = "attendee"
      hash = raids[raidid].events[event].attendees or {}
      raids[raidid].events[event].attendees = hash
   end   
   if self.hasInfo then
      self:info("Adding %s as %s for event %s in raid #%d",
		val, type, event, raidid)
   end
   hash[ val ] = true
   mod:ScheduleRaidChangeData(raids[raidid])
end

function mod:DelExtraAttendee(var,val)
   local type = var[#var]
   local raidid = tonumber(var[#var-2])
   local event = var[#var-1]
   local hash
   val = cap(val)
   if type == "delStandbyInput" then
      type = "standby"
      hash = raids[raidid].events[event].standby
   else
      type = "attendee"
      hash = raids[raidid].events[event].attendees
   end   

   if not hash or not hash[val] then
      if self.hasInfo then self:info("%s is not in the %s list for %s in raid #%d", val, type, event, raidid) end
      return
   end
   hash[val] = nil
   if self.hasInfo then self:info("Deleting %s from the %s list for %s in raid #%d", val, type, event, raidid) end
   mod:ScheduleRaidChangeData(raids[raidid])
end


function mod:BuildRaidConfig(var)
   local raidid = (type(var) == "table" and var[#var-1]) or var
   local raidHash = options.args.raids.args[raidid]
   
   raidHash.args.loader.hidden = true
      
   for event,data in pairs(raids[tonumber(raidid)].events) do
      if self.hasSpam then self:spam("  Added event %s...", event) end
      if not raidHash.args[event] then 
	 raidHash.args[event] = {
	    type = "group",
	    name = event,
	    get = "GetEventOption",
	    set = "SetEventOption",
	    order = data.eventTime,
	    args = {
	    },
	    plugins = {
	       participants = attendeeConfig,
	    }
	 }
      end
      for key, val in pairs(standardEventOptions) do
	 raidHash.args[event].args[key] = val
      end

      if data.items then
	 local subhash = raidHash.args[event].args
	 local looterHash = {
	    ["00BANK"] = L["Bank"],
	    ["00DE"] = L["Disenchant"],
	 }
	 for val in pairs(data.attendees) do
	    looterHash[val] = val
	 end
	 
	 for id,val in pairs(data.items) do
	    local order = 10000 + id*50
	    subhash["header:"..id] = {
	       type = "header",
	       name = "",
	       order = order+4
	    }
	    subhash["name:"..val.itemid..": "..id] = {
	       type = "input",
	       name = GetItemTexture,
	       dialogControl = "MDKPItemWidget",
	       width = "full",
	       get = GetItemDesc,
	       order = order +1 
	    }
	    subhash["looter:"..id] = {
	       type = "select",
	       name = L["Looter"],
	       order = order +2,
	       values = looterHash,
	       get = "GetLootOption",
	       set = "SetLootOption",
	    }
	    subhash["value:"..id] = {
	       type = "input",
	       name = L["Cost"],
	       order = order + 3,
	       get = "GetLootOption",
	       set = "SetLootOption",
	    }
	    --	 subhash["count:"..id] = {
	    --	    type = "range",
	    --	    name = L["Count"],
	    --	    order = order + 3,
	    --	    min = 1, max = 5, step = 1,
	    --	    get = "GetLootOption",
	    --	    set = "SetLootOption",
	    --	 }
	 end
      end
   end
   self:NotifyChange()
end

function mod:RegisterModuleOptions(module, moduleOptions)
   options.args.settings.args[module] = moduleOptions
   self:NotifyChange()
end

function mod:BuildConfiguration()
   local subopts

   UpdateUpvalues()


   -- Custom events
   subopts = options.args.events.args
   for id, event in pairs(events) do
      subopts[id] = stdEventOptions
      cmdoptions.args.event.args[id] = cmdEventOption
   end
   
   
   -- Set up raid zones
   subopts = options.args.zones.args
   for zone, data in pairs(zones) do
      if not subopts[zone] then 
	 local name = data.name 
	 if name == "World Boss" then
	    name = L["World Boss"]
	 else
	    name = ZL[name]
	 end
	 subopts[zone] = {
	    type = "group",
	    name = name,
	    set = "SetZoneOption",
	    get = "GetZoneOption",
	    args = {
	       loader = {
		  name = "Loader",
		  type = "toggle",
		  get = "BuildZoneConfig", 
		  set = function() end,
	       },
	       multiDKP = {
		  name = L["MultiDKP Pool"],
		  type = "select",
		  desc = L["The DKP Pool to use when taking bids and charging for items in this zone. This overrides the default setting but not item specific settings. The default pool will be used if this is unset."],
		  hidden = "MultiDKPDisabled",
		  values = "GetMultiDKPPoolList",
		  order = 1
	       },

	    }
	 }
      end
   end

   for raidid = 1, #raids do
      self:AddRaidConfig(raidid)
   end

   for item,_ in pairs(items) do
      self:AddItemConfig(item)
   end
   -- Profile handling
   options.args.settings.args.profile = DBOpt:GetOptionsTable(mod.db)

   -- setup pplayer database
   mod:BuildPlayerConfig()
end

function mod:DeleteRaidConfig(raidid)
   options.args.raids.args[tostring(raidid)] = nil
end

function mod:AddRaidConfig(raidid)
   local subopts = options.args.raids.args
   local raidkey = tostring(raidid)
   if not subopts[raidkey] or subopts[raidkey].args.loader.hidden then
      subopts[raidkey] = {
	 type = "group",
	 name = GetRaidName(raidid),
	 order = -raidid,
	 args = {
	    loader = {
	       name = "Loader",
	       type = "toggle",
	       get = "BuildRaidConfig", 
	       set = function() end
	    },
	 }
      }
      for key,val in pairs(standardRaidOptions) do
	 subopts[raidkey].args[key] = val
      end
   end
end

function mod:ToggleConfigDialog()
   if C.OpenFrames["Magic DKP"] then
      C:Close("Magic DKP")
   else
      C:Open("Magic DKP")
      self:SetStatusText(fmt(L["Active profile: %s"], self.db:GetCurrentProfile()))
   end
end

do
   local updateStatusTimer
   function mod:SetStatusText(text, update)
      local frame = C.OpenFrames["Magic DKP"]
      if frame then
	 frame:SetStatusText(text)
	 if updateStatusTimer then self:CancelTimer(updateStatusTimer, true) end
	 if update then
	    updateStatustimer = self:ScheduleTimer("SetStatusText", 10, fmt(L["Active profile: %s"], self.db:GetCurrentProfile()))
	 else
	    updateStatustimer = false 
	 end
      end
   end
end

function mod:NotifyChange()
   UpdateUpvalues()
   R:NotifyChange(L["Magic DKP"])
end
 
function mod:SetProfileParam(var, value)
   local varName = var[#var]
   db[varName] = value
   if self.hasSpam then 
      self:spam("Setting parameter %s to %s.", varName, tostring(value))
   end

   if varName == "logLevel" then
      self:SetLogLevel(value)
   elseif varName == "raidsize" then
      self:StartOrEndRaid()
   elseif varName == "xloothook" then
      mod:LoadXLootHook()
   end
end

function mod:GetProfileParam(var) 
   local varName = var[#var]
   if self.hasSpam then
      self:spam("Getting parameter %s as %s.", varName, tostring(db[varName]))
   end
   return db[varName]
end

function mod:UpdateItem(itemid, ignore, bank, value, type)
   itemid = tonumber(itemid)
   local item = items[itemid] or {} 
   item.ignore = ignore -- entirely ignore this item
   item.bank   = bank   -- always send to bank
   item.value  = value  -- default / fixed price
   item.type   = type   -- nil = ask, true = fixed
   item.itemid = itemid
   item.itemInfo = { GetItemInfo(itemid) }
 
   items[itemid] = item

   self:AddItemConfig(itemid)
   self:NotifyChange()
end

function mod:AddItemConfig(itemid)
   local ii = items[itemid].itemInfo
   local itemName = ii[1]
   local itemLink  = ii[2]
   local itemTexture = ii[10]
   local itemRarity = ii[3]
   if not GetItemInfo(itemid) then
      --      if self.hasTrace then self:trace("Unknown item %d, loading via GameTooltip...", itemid) end
      items[itemid].name = items[itemid].name or itemName or "Unknown Item: "..itemid
      items[itemid].rarity = items[itemid].rarity or itemRarity or 100
      items[itemid].unknown = true
      if not itemLink then
	 itemLink = items[itemid].name
      end
   else
      items[itemid].name = itemName
      items[itemid].unknown = nil
      items[itemid].rarity = itemRarity
      items[itemid].texture = itemTexture
   end
   
   items[itemid].multiDKP = "D"
   items[itemid].ilevel = L["Item Level"].." "..(ii[4] or 0)

   itemLink = gsub(gsub(itemLink, "[\[]", ""), "[\]]", "")
   local key = tostring(ii[4] or 0)
   local groupcfg = options.args.itemdb.args[key]
   if not groupcfg then
      groupcfg = {
	 type = "group",
	 name = items[itemid].ilevel,
	 order = 10000-ii[4],
	 args = {}
      }
      options.args.itemdb.args[key] = groupcfg
   end
   
   
   local itemcfg = groupcfg.args[tostring(itemid)] or {}

   itemcfg.type = "group"
   itemcfg.name = itemLink
   itemcfg.order = (20 - items[itemid].rarity) * 1000 + (itemLevel or 0)
   itemcfg.get = "GetOption"
   itemcfg.set = "SetOption"
   itemcfg.args = standardItemOptions
   
   groupcfg.args[tostring(itemid)] = itemcfg
end


function mod:IsDefaultBossEvent(info)
   local mob, region  = tonumber(info[#info-1]), info[#info-2]
   local hash = zones[region].mobs[mob]
   local defHash =  mod.defaultZones[zones[region].name]
   if not defHash or not defHash.mobs[mob] then
      return not hash.event or hash.event == hash.name
   else
      return hash.event == defHash.mobs[mob].event
   end
end

function mod:ResetBossEvent(info, value)
   local mob, region  = tonumber(info[#info-1]), info[#info-2]
   local defHash = mod.defaultZones[zones[region].name]
   if not defHash or not defHash.mobs[mob] then
      zones[region].mobs[mob].event = nil
   else
      zones[region].mobs[mob].event = defHash.mobs[mob].event
   end
end

function mod:SetEventOption(info, value)
   local var = info[#info]
   local event = info[#info-1]
   local raid = tonumber(info[#info-2])
   if var == "value" then
      raids[raid].events[event].value = tonumber(value)
   end
end

function mod:GetEventOption(info)
   local var = info[#info]
   local event = info[#info-1]
   local raid = tonumber(info[#info-2])
   local value = raids[raid].events[event][var]
   if value then 
      return tostring(value)
   elseif var == "value" then
      return "0"
   end
end

function mod:SetLootOption(info, value)
   local var,num = string.match(info[#info], idmatch)
   local event = info[#info-1]
   local raid = tonumber(info[#info-2])
   num = tonumber(num)

   if self.hasSpam then self:spam("|cffffff00SetLootOption:|r %d/%s/%d/%s => %s", raid, event, num, tostring(var), tostring(value)) end

   if var ~= "looter" then
      value = tonumber(value)
   elseif value == "00BANK" then
      value = "bank"
   elseif value == "00DE" then
      value = "disenchant"
   end

   local itemdata =  raids[raid].events[event].items[num]
      
   if var == "looter" and value ~= itemdata.looter then
      -- changing the looter
      mod:AdjustExternalDKP(value, itemdata.value, itemdata.multidkp)
      mod:AdjustExternalDKP(itemdata.looter, -itemdata.value)
   elseif var == "value" and value ~= itemdata.value then
      mod:AdjustExternalDKP(itemdata.looter, value - itemdata.value, itemdata.multidkp)
   end
   itemdata[var] = value
   mod:ScheduleRaidChangeData(raids[raid]) -- send an update if you're the master 
end

function mod:GetLootOption(info)
   local var,num = string.match(info[#info], idmatch)
   local event = info[#info-1]
   local raid = tonumber(info[#info-2])
   num = tonumber(num)
   local value = raids[raid].events[event].items[num][var]
   if self.hasSpam then self:spam("GetLootOption: %d/%s/%d/%s => %s", raid, event, num, tostring(var), tostring(value)) end
   if value and var == "value" then 
      return tostring(value)
   elseif var == "looter" then
      if value == "bank" then
	 value = "00BANK"
      elseif value == "disenchant" then
	 value = "00DE"
      end
   end
   return value
end

function mod:SetZoneOption(info, value)
   local var = info[#info]
   local region = info[#info-1]
   zones[region][var] = value
end

function mod:GetZoneOption(info)
   local var = info[#info]
   local region = info[#info-1]
   local value = zones[region][var]
   if var == "multiDKP" and not value then
      value = "D"
      zones[region][var] = value
   end
   return value
end

function mod:SetBossOption(info, value, state)
   local var = info[#info]
   local mob = tonumber(info[#info-1])
   local region = info[#info-2]
   local old = zones[region].mobs[mob][var]
   if var == "event" and old then
      for id, data in pairs(zones[region].mobs) do
	 if data.event == old then
	    data.event = value
	    if self.hasSpam then
	       self:spam("|cffffff00SetBossOption:|r %s/%d/%s => %s", region, id, var, tostring(value))
	    end
	 end
      end
   else 
      zones[region].mobs[mob][var] = value or nil
   end
   if self.hasSpam then self:spam("|cffffff00SetBossOption:|r %s/%d/%s => %s", region, mob, var, tostring(value)) end
end

function mod:GetBossOption(info, key)
   local var = info[#info]
   local mob = tonumber(info[#info-1])
   local region = info[#info-2]
   local value = zones[region].mobs[mob][var]

   if var == "event" then
      if not value then
	 value  = zones[region].mobs[mob].name
      end
   end
--      value = PRI_LIST[value or 1]
--   elseif var == "category" then
--      value = ACT_LIST[value or 1]
   if self.hasSpam then self:spam("GetBossOption: %s/%d/%s => %s", region, mob, var, tostring(value)) end
   return value 
end

function mod:IsItemIgnored(var)
   return items[tonumber(var[#var-1])].ignore
end

function mod:IsItemUnvalued(var)
   local item = items[tonumber(var[#var-1])]
   return item.ignore or item.bank
end

function mod:IsItemKnown(var)
   local item = items[tonumber(var[#var-1])]
   return not item.unknown
end

function mod:RefreshItemData(itemid)
   local name = GetItemData(itemid)
   if not name then
      self:ScheduleTimer("RefreshItemData", 0.5, tonumber(itemid))
      return
   end
   self:AddItemConfig(itemid)
   self:NotifyChange()
end
				  
function mod:ExpireOOGPlayers()
   mod:RegisterEvent("GUILD_ROSTER_UPDATE", "GUILD_ROSTER_UPDATE_OOG_CLEANUP")
   GuildRoster()
end

function mod:GUILD_ROSTER_UPDATE_OOG_CLEANUP()
   local knownPlayers = {}
   mod:UnregisterEvent("GUILD_ROSTER_UPDATE") 
   local numGuildMembers = GetNumGuildMembers(true)
   for i = 1, numGuildMembers do
      knownPlayers[GetGuildRosterInfo(i)] = true
   end

   for _,data in pairs(raids) do
      for _,event in pairs(data.events) do
	 if event.attendees then
	    for name in pairs(event.attendees) do
	       knownPlayers[name] = true
	    end
	 end
	 if event.standby then
	    for name in pairs(event.standby) do
	       knownPlayers[name] = true
	    end
	 end
      end
   end
   for name,data in pairs(players) do
      if knownPlayers[name] ~= true and
	 (not data.main or knownPlayers[data.main] ~= true) then
	 mod:trace("Deleting old player: "..name)
	 players[name] = nil
	 options.args.playerdb.args[name] = nil
      end
   end
end

function mod:ExpireOldItems()
   local usedItems = {}
   for id in pairs(mod.defaultItems) do
      usedItems[id] = true
   end
   for _,data in pairs(raids) do
      for _,event in pairs(data.events) do
	 if event.items then 
	    for _,itemData in pairs(event.items) do
	       usedItems[itemData.itemid] = true
	    end
	 end
      end
   end
   local deleted = 0
   for id,data in pairs(items) do
      if usedItems[id] ~= true then
	 if not db.deleteBank and data.bank then
	    mod:trace("Not deleting unused item (bank): "..data.name)
	 elseif not db.deleteIgnored and data.ignore then
	    mod:trace("Not deleting unused item (ignore): "..data.name)
	 else
	    mod:trace("Deleting unused item: "..data.name)
	    items[id] = nil
	    deleted = deleted + 1
	 end
      end
   end
   options.args.itemdb.args = {}
   for item,_ in pairs(items) do
      self:AddItemConfig(item)
   end   
end

function mod:ExpireOldRaids()
   local oldestTime = time()
   if db.maxraidage == 1 then
      mod:warn("Not expiring raids since max raid age is set to never delete.")
      return
   end
   if db.maxraidage == 2 then
      oldestTime = oldestTime - 183 * 24 * 3600
   elseif db.maxraidage == 3 then
      oldestTime = oldestTime - 91 * 24 * 3600
   elseif db.maxraidage == 4 then
      oldestTime = oldestTime - 31 * 24 * 3600
   end
   local newRaids = {}
   for id,data in pairs(raids) do
      if data.endTime < oldestTime then
	 mod:debug("Expiring raid: "..GetRaidName(id))
      else
	 newRaids[#newRaids+1] = data
      end      
   end

   mod:info("Remaining raids: %d, Expired Raids: %d", #newRaids, #raids - #newRaids)

   MagicDKPDB.oldRaids = nil
   MagicDKPDB.raids = newRaids
   raids = newRaids
   mod:RefreshRaidConfig()
end

function mod:RefreshRaidConfig()
   options.args.raids.args = {}   
   for raidid = 1, #raids do
      self:AddRaidConfig(raidid)
   end
end

function mod:FetchItemData(var)
   local itemid = var[#var-1]
   GameTooltip:SetHyperlink("item:"..itemid)
   self:ScheduleTimer("RefreshItemData", 0.5, tonumber(itemid))
end

function mod:SetOption(info, value, state)
   local var = info[#info]
   local item = info[#info-1]
   local section = info[#info-3]

   if section == "itemdb" then
      if var == "value" then
	 items[tonumber(item)][var] = value and tonumber(value) or nil
      else
	 items[tonumber(item)][var] = value or nil
      end
   end
   if self.hasSpam then self:spam("|cffffff00SetOption:|r %s/%s/%s => %s", section, item, var, tostring(value)) end
end

function mod:GetEventName(info)
   return events[info[#info-1]]
end

function mod:SetEventName(info, value)
   events[info[#info-1]] = value
end

function mod:GetOption(info, key)
   local var = info[#info]
   local item  = info[#info-1]
   local section = info[#info-3]
   local value
   if section == "itemdb" then
      value = items[tonumber(item)][var]
      if var == "value" then
	 value = tostring(value or 0)
      elseif var == "multiDKP" and not value then
	 items[tonumber(item)][var] = "D"
	 value = "D"
      end
   end
   if self.hasSpam then self:spam("GetOption: %s/%s/%s => %s", section, item, var, tostring(value)) end
   return value or false
end

function mod:AddNewBoss(uid, name, zone)
   zone.mobs[uid] = {
      name = BR[name] or name
   }
   self:BuildZoneConfig(self:MakeKey(zone.name))
   if self.hasInfo then self:info("Detected new boss %s in zone %s.", name, zone.name) end
end

function mod:GetOptions()
   return options
end

function mod:GetZoneName(zone)
   local simple 
   if not zone then
      zone = GetRealZoneText()
   end
   zone = ZR[zone] or zone
   simple = self:MakeKey(zone)
   return simple, zone
end

LibStub("AceConfig-3.0"):RegisterOptionsTable(L["Magic DKP"],  function(name) return (name == "dialog" and options) or cmdoptions end,
					      { "magicdkp", "md", "mdkp" })


------------------------------------------------------------------
-- Raid Viewer UI Below this line...
------------------------------------------------------------------

local RU 
function mod:RU_Setup()
   
end



 
